@@ -18,12 +18,19 @@
 # Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 # 02110-1301, USA.
 
+from __future__ import print_function
+
+import sys
 import os
 import tempfile
 import atexit
 import logging
-import urlparse
-import urllib2
+try:
+    from urllib.parse import urljoin
+    from urllib.request import Request, URLError, urlopen
+except ImportError:
+    from urlparse import urljoin
+    from urllib2 import Request, URLError, urlopen
 import shutil
 import re
 import subprocess
@@ -49,11 +56,19 @@ _bzr_cache_dir = None
 
 
 class AtomicFile(object):
-    """Facilitate atomic writing of files."""
+    """Facilitate atomic writing of files.  Forces UTF-8 encoding."""
 
     def __init__(self, filename):
         self.filename = filename
-        self.fd = open('%s.new' % self.filename, 'w')
+        if sys.version_info[0] < 3:
+            self.fd = codecs.open(
+                '%s.new' % self.filename, 'w', 'UTF-8', 'replace')
+        else:
+            # io.open is available from Python 2.6, but we only use it with
+            # Python 3 because it raises exceptions when passed bytes.
+            self.fd = io.open(
+                '%s.new' % self.filename, mode='w',
+                encoding='UTF-8', errors='replace')
 
     def __enter__(self):
         return self.fd
@@ -68,14 +83,6 @@ class AtomicFile(object):
         self.fd.write(s)
 
 
-class AtomicUTF8File(AtomicFile):
-    """As AtomicFile, but forces UTF-8 encoding."""
-
-    def __init__(self, filename):
-        self.filename = filename
-        self.fd = codecs.open('%s.new' % self.filename, 'w', 'utf8', 'replace')
-
-
 class SeedError(RuntimeError):
     """An error opening or parsing a seed."""
 
@@ -115,16 +122,16 @@ class Seed(object):
                                     "  '%s'" % (status, ' '.join(command)))
             return open(os.path.join(checkout, name))
         else:
-            url = urlparse.urljoin(path, name)
+            url = urljoin(path, name)
             _logger.info("Downloading %s", url)
-            req = urllib2.Request(url)
+            req = Request(url)
             req.add_header('Cache-Control', 'no-cache')
             req.add_header('Pragma', 'no-cache')
-            return urllib2.urlopen(req)
+            return urlopen(req)
 
     def __init__(self, bases, branches, name, bzr=False):
         """Read a seed from a collection."""
-        if isinstance(branches, str) or isinstance(branches, unicode):
+        if isinstance(branches, basestring):
             branches = [branches]
 
         self._name = name
@@ -146,7 +153,7 @@ class Seed(object):
                         r'bzr\+ssh://(?:[^/]*?@)?(.*?)(?:/|$)', base)
                     if ssh_match:
                         ssh_host = ssh_match.group(1)
-                except (OSError, IOError, urllib2.URLError):
+                except (OSError, IOError, URLError):
                     pass
             if fd is not None:
                 break
@@ -174,18 +181,23 @@ class Seed(object):
                         path = os.path.join(base, branch)
                         if not path.endswith('/'):
                             path += '/'
-                        _logger.warning(
-                            '  %s' % urlparse.urljoin(path, name))
+                        _logger.warning('  %s' % urljoin(path, name))
             raise SeedError("Could not open %s" % name)
 
         try:
             self._text = fd.read()
+            # In Python 3, we need to decode seed text read from URLs.
+            if sys.version_info[0] >= 3 and isinstance(self._text, bytes):
+                self._text = self._text.decode(errors="replace")
         finally:
             fd.close()
 
     def open(self):
         """Open a file object with the text of this seed."""
-        self._file = io.BytesIO(self._text)
+        if sys.version_info[0] < 3:
+            self._file = io.BytesIO(self._text)
+        else:
+            self._file = io.StringIO(self._text)
         return self._file
 
     def read(self, *args, **kwargs):
@@ -236,12 +248,37 @@ class Seed(object):
         """The text of this seed."""
         return self._text
 
-    def __cmp__(self, other):
-        """Compare two seeds."""
-        if isinstance(other, Seed):
-            return cmp(self.text, other.text)
-        else:
-            return cmp(self.text, other)
+    def __lt__(self, other):
+        if not isinstance(other, Seed):
+            return NotImplemented
+        return self.text < other.text
+
+    def __le__(self, other):
+        if not isinstance(other, Seed):
+            return NotImplemented
+        return self.text <= other.text
+
+    def __eq__(self, other):
+        if not isinstance(other, Seed):
+            return NotImplemented
+        return self.text == other.text
+
+    def __ne__(self, other):
+        if not isinstance(other, Seed):
+            return NotImplemented
+        return self.text != other.text
+
+    def __ge__(self, other):
+        if not isinstance(other, Seed):
+            return NotImplemented
+        return self.text >= other.text
+
+    def __gt__(self, other):
+        if not isinstance(other, Seed):
+            return NotImplemented
+        return self.text > other.text
+
+    __hash__ = None
 
 
 class CustomSeed(Seed):
@@ -482,25 +519,26 @@ class SeedStructure(collections.Mapping, object):
         """Write the text of the seed STRUCTURE file."""
         with AtomicFile(filename) as f:
             for line in self._lines:
-                print >>f, line
+                print(line, file=f)
 
     def write_dot(self, filename):
         """Write a dot file representing this structure."""
-        with AtomicUTF8File(filename) as dotfile:
-            print >>dotfile, "digraph structure {"
-            print >>dotfile, "    node [color=lightblue2, style=filled];"
+        with AtomicFile(filename) as dotfile:
+            print("digraph structure {", file=dotfile)
+            print("    node [color=lightblue2, style=filled];", file=dotfile)
 
             for seed in self._seed_order:
                 if seed not in self._original_inherit:
                     continue
                 for inherit in self._original_inherit[seed]:
-                    print >>dotfile, "    \"%s\" -> \"%s\";" % (inherit, seed)
+                    print("    \"%s\" -> \"%s\";" % (inherit, seed),
+                          file=dotfile)
 
-            print >>dotfile, "}"
+            print("}", file=dotfile)
 
     def write_seed_text(self, filename, seedname):
         """Write the text of a seed in this collection."""
         with AtomicFile(filename) as f:
             with self._seeds[seedname] as seed:
                 for line in seed:
-                    print >>f, line.rstrip('\n')
+                    print(line.rstrip('\n'), file=f)
